//-----------------------------------------------------------------------------------
// FILE NAME: terminalPC.c
// TARGET DEVICE: 	LPC2103
// PROGRAMMER DEVICE:	GNU/Linux
// CREATED ON:          21/08/2017
// CREATED BY: 		JESUS GARCIA BARAJAS
//
// Code to program a C8051Fxxx device through its C2 interface.
//
// This software targets a PC device, though it could be easily
// ported to any device.
//------------------------------------------------------------------------------------
/**********************************************************
                  		Header files
 **********************************************************/
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <asm/termios.h>
#include <fcntl.h>
#include <stdio.h>

/**********************************************************
                       Define SECTION
**********************************************************/
#define DEV "/dev/ttyUSB0"
#define BAUDRATE B115200
/**********************************************************
                      FUNTION SECTION
**********************************************************/

//-----------------------------------------------------------------------------------
// safewrite()
//-----------------------------------------------------------------------------------
// - Retry until you write the entire buffer
//-----------------------------------------------------------------------------------
// -------------------------------------------------------
// write seguro. 
void safewrite(int fd,char *buf,int n)
{
  int i,j;
  i=0;
  do {
    j=write(fd,&buf[i],n);
    usleep(6000);
    i+=j; n-=j;
  } while (n);
}
//-----------------------------------------------------------------------------------
// writeSerial()
//-----------------------------------------------------------------------------------
// - Write buffer
//-----------------------------------------------------------------------------------
int writeSerial( int fd, unsigned int *buffer, int ilength )
{
    int written;
    int result;
    int i;
    int o;
    int length;
    unsigned int out[256];


    printf("writeSerial ilength = %d\n", ilength);


    o = 0;
    for( i = 0; i < ilength; i++ )
                      {
                        out[o++] = buffer[i];
                      }
    out[o++] = '\r';
    out[o++] = '\n';

    length = o;

    written = 0;
    result = 0;

    while( length - written > 0 ){
    
                /*
                printf("writeSerial length = %d: ", length - written);//DEBUG
                for( i = 0; i < length - written; i++ ){
  
                                  printf("%02x", * (out + written + i) );
                }
                printf("\n");
                */
                result = write( fd, out + written, length - written );
                if ( result >= 0 ){
                          
                            written = written + result;
                }
                else
                {
                break;
                }
    }

    if ( result < 0 )
    {
      perror( "writing SERIAL failed:");
      goto exit;
    }

    result = 0;

    exit:
    return result;
}

//-----------------------------------------------------------------------------------
// ascii2hex( )
//-----------------------------------------------------------------------------------
// - Pass a character to its hex value
//-----------------------------------------------------------------------------------
static int ascii2hex( char c )

{
    int ret = -1;

    if ( ( c >= '0' ) && ( c <= '9' ) )
    {
  ret = c - '0';
    }
    else if ( ( c >= 'A' ) && ( c <= 'Z' ) )
    {
  ret = c - 'A' + 10;
    }
    else if ( ( c >= 'a') && ( c <= 'z' ) )
    {
  ret = c - 'a' + 10;
    }
    return ret;
}

//-----------------------------------------------------------------------------------
// TransStrHex( )
//-----------------------------------------------------------------------------------
// - Pass a character from a file.hex to its hexadecimal value
//-----------------------------------------------------------------------------------
int TransStrHex(unsigned char a)
{
  switch (a)
  {
    case '0':
    return 0b0000;
    case '1':
    return 0b0001;
    case '2':
    return 0b0010;
    case '3':
    return 0b0011;
    case '4':
    return 0b0100;
    case '5':
    return 0b0101;
    case '6':
    return 0b0110;
    case '7':
    return 0b0111;
    case '8':
    return 0b1000;
    case '9':
    return 0b1001;
    case 'A':
    return 0b1010;
    case 'B':
    return 0b1011;
    case 'C':
    return 0b1100;
    case 'D':
    return 0b1101;
    case 'E':
    return 0b1110;
    case 'F':
    return 0b1111;
    case ':':
    return 0x3A;
    case 10:
    return 0x0A;
    case 13:
    return 0x0D;
    
    default:
    printf("Dato erroneo, elimina los comentarios del archivo .hex\n");
  }
}
/**********************************************************
                      MAIN SECTION
**********************************************************/

main(int argc,char **argv)
{

	int i,j,fd,fdin,baud=115200,result,hex=0,crlf=0,sl;
	struct termios term;

	int y,w,z,h;
    unsigned int Hex,Num,Address=0,Type,AddressExt,Comand,SFR,NewSFR,NumPage; 

    char Fichero[20];
    char *ptr=Fichero;
    unsigned int BloqueReadAux[256];
    unsigned int BloqueWrite[256][256]; 
    char BloqueRead[256];
    //unsigned int *PTR_Write=BloqueWrite;
    char *PTR_Read=BloqueRead;
    for(i=0;i<(256);i++){
                      BloqueReadAux[i]=0x00; 
                      BloqueRead[i]=0x00; 
                      for(y=0;y<(256);y++) {
                        	BloqueWrite[i][y]=0x00; 
                      }
    }
	unsigned char a,*dn=DEV,fn[256];
        const int baudtable[][2]={
            50,B50,
            75,B75,
            110,B110,
            150,B150,
            200,B200,
            300,B300,
            600,B600,
            1200,B1200,
            2400,B2400,
            4800,B2400,
            9600,B9600,
            19200,B19200,
            38400,B38400,
	    	    57600,B57600,
            115200,B115200,
	    	    921600,B921600,
	    	    1500000,B1500000,
            0,0
        };
	

        
/********************************* Program menu **************************************/
    for (y=1;y<argc;y++){
            if (*argv[y]=='-'){
                switch(argv[y][1]){ 
                case 'h':   printf("Uso: %s opciones\n",argv[0]);
                            printf("Opciones:\n");
                            printf("-h : Help\n");
                            printf("-w <archivo.hex> : copiar archivo.hex en Flash\n");
                            printf("-b <archivo.bin> : copiar archivo.bin en Flash\n");
                            printf("-r <Direccion dec> <Numero de Bytes>: Leer Flash \n");
			                printf("-d <Numero de pagina> : Borra dicha pagina de la Flash\n");
                            printf("-e : Borra dispositivo \n");
                            printf("-s <Direccion SFR hex> <Nuevo SFR hex>: Cambia el SFR\n");
                            printf("-l <Direccion SFR hex> : Lee el SFR\n");
			                exit(0);
                	    case 'w':   Comand=0; ptr=argv[++y]; break;
                	    case 'b':   Comand=1; ptr=argv[++y]; break;
		            case 'r':   Comand=2; Address=atoi(argv[++y]);NumPage=atoi(argv[++y]); break;
		            case 'd':   Comand=3; NumPage=atoi(argv[++y]); break;
		            case 'e':   Comand=4; break;
		            case 's':   Comand=5; SFR=((TransStrHex(argv[2][0]))<<4)|TransStrHex(argv[2][1]); 
		        			          NewSFR=((TransStrHex(argv[3][0])<<4))|TransStrHex(argv[3][1]); break;
		            case 'l':   Comand=6; SFR=((TransStrHex(argv[2][0]))<<4)|TransStrHex(argv[2][1]); break;
                }
        	}    
	}
                                  // Debug program menu
    printf("Su Eleccion:\n");
    printf("Comando:%d\n",Comand);
    printf("Fichero:[%s]\n",ptr);
    printf("Numero pagina:%d\n",NumPage);
    printf("SFR:%02x\n",SFR);
    printf("Nuevo SFR:%02x\n",NewSFR);

/************************** Connection configuration *******************************/

        for (i=0;baudtable[i][0];i++) {
            if (baud==baudtable[i][0]) break;
        }
        if (!baudtable[i][0]) {printf("Velocidad no soportada\n"); exit(0);}

        printf("Puerto serie: %s\n",dn);//DEBUG
        printf("Velocidad: %d baudios\n",baud);//DEBUG


	// Open port 
	if ((fd=open(dn,O_RDWR/*|O_NONBLOCK*/))==-1) {perror("open"); exit(0);}
	bzero(&term,sizeof(term));
	term.c_cflag = baudtable[i][1] | CS8 | CLOCAL | CREAD;
	term.c_cc[VMIN]=1;	// blocking mode
	if (tcsetattr(fd,TCSANOW,&term)) { fprintf(stderr,"Fallo en tcsetattr\n"); exit(0); }
	if (tcflush(fd,TCIFLUSH)) { fprintf(stderr,"Fallo en tcflush\n"); exit(0); }

	//  STDIN  
	i=fcntl(0,F_GETFL); fcntl(0,F_SETFL,i|O_NONBLOCK);
	tcgetattr(0,&term);

	term.c_lflag&=~ICANON;
	term.c_lflag&=~ECHO;

	if (tcsetattr(0,TCSANOW,&term)) {
		fprintf(stderr,"Fallo en tcsetattr\n");
		exit(0);
	}
	
	// Reset
    if(ioctl(fd, TIOCMGET, &i)) {fprintf(stderr,"Fallo en ioctl TIOCMGET\n"); exit(0);}
    i|=TIOCM_DTR|TIOCM_RTS; // Reset y Bootloader L
    if(ioctl(fd, TIOCMSET, &i)) {fprintf(stderr,"Fallo en ioctl TIOCMSET\n"); exit(0);}
    usleep(300000);

	// Start
    i&=~TIOCM_RTS;        // bootloader H
    if(ioctl(fd, TIOCMSET, &i)) {fprintf(stderr,"Fallo en ioctl TIOCMSET\n"); exit(0);}
    usleep(300000);
    i&=~TIOCM_DTR;        // Reset H
    if(ioctl(fd, TIOCMSET, &i)) {fprintf(stderr,"Fallo en ioctl TIOCMSET\n"); exit(0);}

/***************************** IN/OUT DATA **************************************************/
   	
	fdin=0;
	hex=1;
	crlf=0;
//	for (;;) { //for Debug
		sl=2;
		if (read(fd,&a,1)==1) {
			if (hex) {printf("%02X ",a); fflush(stdout);}
			else write(1,&a,1);
		} else sl--;
		j=read(fdin,&a,1);
		if (j==0 && fdin) {
			close(fdin);
			fdin=0;	
		} 
		if (j==1) {
			switch (a) {//File
			case 	'X'-64:	a=3;//Ctl-X
				break;
			case	'\n':	if (crlf) a='\r';
				break;				
			case	'F'-64:	printf("\nInclude File: ");//Ctl-F
				fflush(stdout);
				tcgetattr(0,&term);
				term.c_lflag|=ECHO;
				tcsetattr(0,TCSANOW,&term);
				i=fcntl(0,F_GETFL); fcntl(0,F_SETFL,i&(~O_NONBLOCK));
				fgets(fn,255,stdin); fn[strlen(fn)-1]=0;
				fcntl(0,F_SETFL,i/*|O_NONBLOCK*/);
				term.c_lflag&=~ECHO;
				tcsetattr(0,TCSANOW,&term);
				if ((fdin=open(fn,O_RDONLY))==-1) fdin=0;
				//continue;
			}
			write(fd,&a,1); usleep(7000);
			if (a=='\r') {a='\n'; write(fd,&a,1); usleep(20000);}
		} else sl--;
		if (!sl) usleep(40000);

//	}// for Debug
  
/***************************** Selection of the chosen command *********************************************/

if (Comand==0){
         FILE *pFichero;
     unsigned char byte;  
     pFichero=fopen(ptr, "r"); 
     if (pFichero==NULL) {
      printf("No puedo abrir el archivo %s\n\n",argv[2]);
      return 0;
      } 
     for(i=0;i<254;i++){// i= line
        w=0;// w= position in the linee
        do {
           
           byte=fgetc(pFichero); 
           
           //printf("dato %d,%d:%c\n",i,j,byte);//DEBUG
           
           BloqueWrite[i][w]=(byte);
        

           w++;
          
            

        }while ((!((BloqueWrite[i][w-1]==10)&&(BloqueWrite[i][w-2]==13)))&&(!(feof(pFichero))));
           
            *PTR_Read=BloqueRead;
            BloqueRead[0]=0;
            
            Num=((TransStrHex(BloqueWrite[i][1]))<<4)|TransStrHex(BloqueWrite[i][2]);
            printf("Num:%d\n",Num );
            BloqueRead[1]=Num;
            Address=((TransStrHex(BloqueWrite[i][3]))<<12)|((TransStrHex(BloqueWrite[i][4]))<<8)|
			((TransStrHex(BloqueWrite[i][5]))<<4)|(TransStrHex(BloqueWrite[i][6]));
            printf("Address:%04x\n",Address );
            BloqueRead[2]=((Address>>8)&0x00FF);
            BloqueRead[3]=(Address&0x00FF);
            Type=((TransStrHex(BloqueWrite[i][7]))<<4)|(TransStrHex(BloqueWrite[i][8]));
            printf("Type:%02x\n",Type);
            
            if (Type==0){
              h=4;
              for(z=0;z<(2*Num);z=z+2){
                          
                          Hex=((TransStrHex(BloqueWrite[i][9+z]))<<4)|(TransStrHex(BloqueWrite[i][10+z]));
                          BloqueReadAux[z]=Hex;
                          //printf("Hex:%x\n",BloqueReadAux[z]);//debug
                          BloqueRead[h++]=BloqueReadAux[z];
              }
	          BloqueRead[h++]='\r';
              BloqueRead[h++]='\n';
              for(z=0;z<30;z++){
              	printf("Hex%d:%02x\n",z,BloqueRead[z]);//debug
              }
              
              safewrite(fd,PTR_Read,(Num+8));
		      
		      usleep(80000);
		      
              do{
                    read(fd,&a,1);
                  }while(a!=0x0C);
              // printf("\r\n\nCopia de fragmento de archivo correcta\r\n\n");//DEBUG
             
            }  
            if (Type==1) {
                            fclose(pFichero);
                            break;
                          }
            if (Type==4) {
                            AddressExt=((TransStrHex(BloqueWrite[i][9]))<<12)|((TransStrHex(BloqueWrite[i][10]))<<8)
					|((TransStrHex(BloqueWrite[i][11]))<<4)|(TransStrHex(BloqueWrite[i][12]));
                            printf("AddressExt:%x\n",AddressExt);
            }               
    }//for
}// if comand=0
/**********************************************************************************************/
if (Comand==1){

    FILE *pFichero;
    unsigned char byte; 
    pFichero=fopen(ptr, "rb"); 
    if (pFichero==NULL) {
      printf("No puedo abrir el archivo %s\n\n",argv[2]);
      return 0;
      }
    for(i=0;i<80;i++){// i=line
        j=0;// j=position in the line
        do {
           
           byte=fgetc(pFichero); 
           
           //printf("dato %d,%d:%x\n",i,j,byte);
           
           BloqueWrite[i][j]=(byte&0x000000FF);
        

           j++;
           
            

        }while ((j<200)&&(!(feof(pFichero))));
           
        Num=j;
        *PTR_Read=BloqueRead;
        
        BloqueRead[0]=1;
        //printf("Num:%d\n",Num );//debug
        BloqueRead[1]=Num;
        
        BloqueRead[2]=((Address>>8)&0x00FF);
        BloqueRead[3]=(Address&0x00FF);
            
        for(z=4;z<(Num+4);z++){
                        BloqueRead[z]=BloqueWrite[i][z];
                        //printf("Hex%d:%02x\n",z,BloqueRead[z]);//DEBUG
              }
        BloqueRead[z++]='\r';
        BloqueRead[z++]='\n';
        for(z=0;z<256;z++){
                        
                        printf("Hex%d:%02x\n",z,BloqueRead[z]);
              }

      
        safewrite(fd,PTR_Read,(Num+6));
        usleep(80000);
 		
              do{
                    read(fd,&a,1);
                  }while(a!=0x0C);
        //printf("\r\n\nCopia de fragmento de archivo correcta\r\n\n");//DEBUG
        Address=Address+Num;
        //printf("Address:%x\n",Address );//DEBUG

        if (feof(pFichero)) {
                        fclose(pFichero); 
                        return 0;
                        } /*feof: check if you get to the end of the file */              
    }
     fclose(pFichero); 

     getchar(); 
     return 0; 



}
            
/**********************************************************************************************/            
if (Comand==2){

	  PTR_Read=BloqueRead;
      BloqueRead[0]=2;
      BloqueRead[1]=NumPage;
      BloqueRead[2]=Address;
      BloqueRead[3]='\r';
      BloqueRead[4]='\n';
      
      safewrite(fd,PTR_Read,5);
      usleep(40000);

                do{
                    read(fd,&a,1);
                  }while(a!=0x0C);
        //printf("\r\n\nLectura  de fragmento de archivo correcta\r\n\n");//DEBUG
 }
            
/**********************************************************************************************/ 
if (Comand==3){

      PTR_Read=BloqueRead;
      BloqueRead[0]=3;
      BloqueRead[1]=NumPage;
      BloqueRead[2]='\r';
      BloqueRead[3]='\n';
      
      safewrite(fd,PTR_Read,4);
      usleep(40000);

                do{
                    read(fd,&a,1);
                  }while(a!=0x0C);
        //printf("\r\n\nBorrado de página correcto\r\n\n");//DEBUG

}// if comand=3
            
/**********************************************************************************************/ 
if (Comand==4){

      PTR_Read=BloqueRead;
      BloqueRead[0]=4;
      BloqueRead[1]='\r';
      BloqueRead[2]='\n';
      
      safewrite(fd,PTR_Read,3);
      usleep(40000);

                do{
                    read(fd,&a,1);
                  }while(a!=0x0C);
        //printf("\r\n\nBorrado de dispositivo correcto\r\n\n");//DEBUG

}
            
/**********************************************************************************************/ 
if (Comand==5){

      PTR_Read=BloqueRead;
      BloqueRead[0]=5;
      BloqueRead[1]=SFR;
      BloqueRead[2]=NewSFR;
      BloqueRead[3]='\r';
      BloqueRead[4]='\n';
     
      safewrite(fd,PTR_Read,5);
      usleep(40000);

                do{
                    read(fd,&a,1);
                  }while(a!=0x0C);
        //printf("\r\n\nEscritura de SFR correcta\r\n\n");//DEBUG

}
            
/**********************************************************************************************/ 
if (Comand==6){

      PTR_Read=BloqueRead;
      BloqueRead[0]=6;
      BloqueRead[1]=SFR;
      BloqueRead[2]='\r';
      BloqueRead[3]='\n';   
      
      safewrite(fd,PTR_Read,5);
      usleep(40000);

                do{
                    read(fd,&a,1);
                  }while(a!=0x0C);
        //printf("\r\n\nLectura de SFR correcta\r\n\n");//DEBUG

}
            
/**********************************************************************************************/          
	         
	      
while (1);//PCON=1;		// CPU idle      
}//last

